#define _XOPEN_SOURCE 600

#include "term.h"
#include "handler.h"
#include "fill.h"
#include "rszbox.h"
#include "key.h"
#include "timer.h"
#include "list.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <sys/wait.h>

#ifdef __FreeBSD__
#include <libutil.h>
#endif

#define UPDATE_FLG	(1<<7)

cui
cui_term_new(cui parent, int x, int y, int w, int h)
{
	cui obj = cui_alloc(sizeof(struct cui_term));
	cui_term_init(obj, parent, x, y, w, h);
	return obj;
}

void
cui_term_init(cui obj, cui parent, int x, int y, int w, int h)
{
	cui_term p = (cui_term)obj;
	cui_base_init(obj, parent, x, y, w, h);

	cui_term_buf_alloc(obj);
	p->cx = p->cy = p->bfy = 0;
	p->cursor_onoff = TRUE;
	p->attr = CUI_ATTR_NORMAL;
	
	p->update_all = FALSE;
	p->esc_buf_n = 0;
	p->ow = obj->w;
	p->oh = obj->h;

	cui_bind(obj, CUI_EVT_DRAW|CUI_EVT_RESIZE, cui_term_hdr, NULL);
}

static void
hdr_update_flush(cui obj, int *attr, char *lp, int x, int y)
{
	cui_term p = (cui_term)obj;

	if(*attr == -1) return;
	*lp = '\0';
	cui_draw_str(obj, x, y, p->lbuf, *attr & ~UPDATE_FLG);
	*attr = -1;
}

static void
hdr_update(cui obj)
{
	cui_term p = (cui_term)obj;
	char *lp = p->lbuf;
	int x, y, attr = -1, lx = 0;

	if(p->update_all) cui_clear(obj);
	for(y=0; y<obj->h; y++){
		for(x=0; x<obj->w; x++){
			char *bp = cui_term_buf(obj, x, y);
			if(p->update_all || (bp[1] & UPDATE_FLG)){
				if(attr != -1 && bp[1] != attr) hdr_update_flush(obj, &attr, lp, lx, y);
				if(attr == -1){
					lp = p->lbuf;
					*lp++ = bp[0] ? bp[0] : ' ';
					attr = bp[1];
					lx = x;
					bp[1] &= ~UPDATE_FLG;
					continue;
				}
				/* bp[1] == attr */
				*lp++ = bp[0] ? bp[0] : ' ';
				bp[1] &= ~UPDATE_FLG;
				continue;
			}
			hdr_update_flush(obj, &attr, lp, lx, y);
		}
		hdr_update_flush(obj, &attr, lp, lx, y);
	}
	if(p->cursor_onoff){
		char buf[2];
		buf[0] = cui_term_buf(obj, p->cx, p->cy)[0];
		buf[1] = '\0';
		if(!buf[0]) buf[0] = ' ';
		cui_draw_str(obj, p->cx, p->cy, buf, CUI_ATTR_REVERSE);
	}
	p->update_all = FALSE;
}

static void
hdr_resize(cui obj)
{
	cui_term p = (cui_term)obj;
	struct cui_term ot;
	cui o = (cui)&ot;
	int x, y;
	char *s, *d;

	ot = *p; /* buf share */
	o->w = ot.ow;
	o->h = ot.oh;

	cui_term_buf_alloc(obj);
	p->ow = obj->w;
	p->oh = obj->h;

	for(y=0; y<obj->h; y++){
		if(y >= o->h) continue;
		for(x=0; x<obj->w; x++){
			if(x >= o->w) continue;
			s = cui_term_buf(o, x, y);
			d = cui_term_buf(obj, x, y);
			d[0] = s[0];
			d[1] = s[1];
		}
	}
	cui_term_buf_free(o);
	
	if(p->cx >= obj->w) p->cx = obj->w - 1;
	if(p->cy >= obj->h) p->cy = obj->h - 1;
}

int
cui_term_hdr(cui obj, int evt, int val, void *prm)
{
	cui_term p = (cui_term)obj;

	switch(evt){
	case CUI_EVT_DRAW:
		p->update_all = TRUE;
		hdr_update(obj);
		return TRUE;
	case CUI_EVT_RESIZE:
		if(p->ow == obj->w && p->oh == obj->h) return TRUE;
		hdr_resize(obj);
		return TRUE;
	}
	return FALSE;
}

void
cui_term_buf_alloc(cui obj)
{
	cui_term p = (cui_term)obj;
	int size = 2 * obj->w * obj->h; /* 2 : char, attr */
	int sz = size + obj->w + 1; /* for lbuf */
	if((p->buf = malloc(sz)) == NULL) ERR("No Mem");
	memset(p->buf, 0, size);
	p->lbuf = p->buf + size;
}

void
cui_term_buf_free(cui obj)
{
	cui_term p = (cui_term)obj;
	if(p->buf){
		free(p->buf);
		p->buf = NULL;
	}
}

void
cui_term_putc(cui obj, char c)
{
	cui_term_putc_buf(obj, c);
	cui_term_update(obj);
}

static int
esc_cmp(cui obj, char *targ)
{
	cui_term p = (cui_term)obj;

	if(strcmp(p->esc_buf, targ) == 0) return 0; /* equal */
	if(strncmp(p->esc_buf, targ, p->esc_buf_n) == 0) return 1; /* continue */
	return -1; /* not equal */
}

static void
esc_flush(cui obj)
{
	cui_term p = (cui_term)obj;
	cui_term_puts_buf_inner(obj, p->esc_buf);
	p->esc_buf_n = 0;
}

static void
esc_do(cui obj, char c)
{
	cui_term p = (cui_term)obj;
	int ret, i, x, y;

	p->esc_buf[ p->esc_buf_n++ ] = c;
	p->esc_buf[ p->esc_buf_n ] = '\0';

	if(p->esc_buf_n == 1) return;
	if(p->esc_buf_n == 2){
		if(c == '[') return;
		esc_flush(obj);
		return;
	}
	if(p->esc_buf_n == 3){
		if(c == '?') return;
		if(isdigit(c)) return;
		esc_flush(obj);
		return;
	}

	/* esc_buf_n >= 4 */
	
	if(p->esc_buf[2] == '?'){
		if((ret = esc_cmp(obj, "\033[?25h")) == 1) return;
		if(ret == 0){
			p->cursor_onoff = TRUE;
			p->esc_buf_n = 0;
			cui_term_update(obj);
			return;
		}
		if((ret = esc_cmp(obj, "\033[?25l")) == 1) return;
		if(ret == 0){
			p->cursor_onoff = FALSE;
			p->esc_buf_n = 0;
			cui_term_update(obj);
			return;
		}
		esc_flush(obj);
		return;
	}

	/* esc_buf[2] is digit */

	if((ret = esc_cmp(obj, "\033[0m")) == 1) return;
	if(ret == 0){
		p->attr = CUI_ATTR_NORMAL;
		p->esc_buf_n = 0;
		return;
	}
	if((ret = esc_cmp(obj, "\033[4m")) == 1) return;
	if(ret == 0){
		p->attr |= CUI_ATTR_ULINE;
		p->esc_buf_n = 0;
		return;
	}
	if((ret = esc_cmp(obj, "\033[7m")) == 1) return;
	if(ret == 0){
		p->attr |= CUI_ATTR_REVERSE;
		p->esc_buf_n = 0;
		return;
	}
	if((ret = esc_cmp(obj, "\033[24m")) == 1) return;
	if(ret == 0){
		p->attr &= ~CUI_ATTR_ULINE;
		p->esc_buf_n = 0;
		return;
	}
	if((ret = esc_cmp(obj, "\033[27m")) == 1) return;
	if(ret == 0){
		p->attr &= ~CUI_ATTR_REVERSE;
		p->esc_buf_n = 0;
		return;
	}

	/* ESC "[(y);(x)H" */
	
	x = y = 0;
	i = 2;
	while(p->esc_buf[i] && isdigit(p->esc_buf[i])){
		y = y * 10 + (p->esc_buf[i++] - '0');
	}
	if(p->esc_buf[i] == '\0') return;
	if(p->esc_buf[i++] != ';'){
		esc_flush(obj);
		return;
	}
	while(p->esc_buf[i] && isdigit(p->esc_buf[i])){
		x = x * 10 + (p->esc_buf[i++] - '0');
	}
	if(p->esc_buf[i] == '\0') return;
	if(p->esc_buf[i++] != 'H'){
		esc_flush(obj);
		return;
	}
	cui_term_move(obj, x-1, y-1);
	p->esc_buf_n = 0;
}

void
cui_term_putc_buf(cui obj, char c)
{
	cui_term p = (cui_term)obj;
	if(c == '\033' || p->esc_buf_n > 0){
		esc_do(obj, c);
		return;
	}
	cui_term_putc_buf_inner(obj, c);
}

void
cui_term_putc_buf_inner(cui obj, char c)
{
	cui_term p = (cui_term)obj;
	int cx, cy;
	char *bp;

	switch(c){
	case '\r':
		cui_term_cr(obj);
		return;
	case '\n':
		cui_term_lf(obj);
		return;
	case '\t':
		cx = (p->cx / 8 + 1) * 8;
		if(cx < obj->w) cui_term_move(obj, cx, p->cy);
		else cui_term_nl(obj);
		return;
	case '\b':
		cx = p->cx;
		cy = p->cy;
		if(cx > 0) cx--;
		else if(cy > 0){
			cy--;
			cx = obj->w - 1;
		}
		cui_term_move(obj, cx, cy);
		return;
	}

	bp = cui_term_buf(obj, p->cx, p->cy);
	bp[0] = c;
	bp[1] = p->attr | UPDATE_FLG;

	if(++p->cx >= obj->w) cui_term_nl(obj);
}

void
cui_term_puts(cui obj, char *s)
{
	cui_term_puts_buf(obj, s);
	cui_term_update(obj);
}

void
cui_term_puts_buf(cui obj, char *s)
{
	while(*s) cui_term_putc_buf(obj, *s++);
}

void
cui_term_puts_buf_inner(cui obj, char *s)
{
	while(*s) cui_term_putc_buf_inner(obj, *s++);
}

void
cui_term_cr(cui obj)
{
	cui_term p = (cui_term)obj;
	if(p->cursor_onoff) cui_term_buf(obj, p->cx, p->cy)[1] |= UPDATE_FLG;
	p->cx = 0;
}

void
cui_term_lf(cui obj)
{
	cui_term p = (cui_term)obj;
	if(p->cursor_onoff) cui_term_buf(obj, p->cx, p->cy)[1] |= UPDATE_FLG;
	if(p->cy < obj->h-1) p->cy++;
	else{
		memset(cui_term_buf(obj, 0, 0), 0, obj->w * 2); /* 2 : char, attr */
		if(++p->bfy >= obj->h) p->bfy = 0;
		p->update_all = TRUE;
	}
}

void
cui_term_nl(cui obj)
{
	cui_term_cr(obj);
	cui_term_lf(obj);
}

char *
cui_term_buf(cui obj, int cx, int cy)
{
	cui_term p = (cui_term)obj;
	int iy = (p->bfy + cy) % obj->h;
	int i = (obj->w * iy + cx) * 2; /* 2 : char, attr */
	return &p->buf[i];
}

void
cui_term_move(cui obj, int cx, int cy)
{
	cui_term p = (cui_term)obj;
	if(p->cursor_onoff) cui_term_buf(obj, p->cx, p->cy)[1] |= UPDATE_FLG;
	p->cx = cx;
	p->cy = cy;
}

void
cui_term_update(cui obj)
{
	hdr_update(obj);
}

/**/

#define TIMEOUT_MSEC	(3*1000)

static cui_list t_list = NULL;

static void
sig_hdr(int sig)
{
	int  pid;
	cui_list p = t_list;
	cui obj;
	cui_terminal t;

	if(sig != SIGCHLD) return;
	pid = wait(NULL);
	for(; p; p=p->next){
		obj = *(cui *)p->data;
		t = (cui_terminal)obj;
		if(t->pid == pid) break;
	}
	if(p == NULL) return;
	cui_readable_del(obj, t->pipe_r);
	cui_list_del(&t_list, p->data);
}

static void
boot_sh(cui obj)
{
	cui_terminal p = (cui_terminal)obj;
	int ret, master;
	signal(SIGCHLD, sig_hdr);
#if defined(__FreeBSD__)
	if((ret = forkpty(&master, NULL, NULL, NULL)) < 0) ERR("forkpy");
#else /* defined(__linux__) */
	int slave;
	char *name;

	if((master = posix_openpt(O_RDWR)) < 0) ERR("posix_openpt");
	if((ret = fork()) < 0) ERR("fork");
#endif
	if(ret != 0){ /* parent */
		p->pipe_r = p->pipe_w = master;
		p->pid = ret;
		*(cui *)cui_list_add(&t_list, sizeof(cui)) = obj;
		return;
	}

	/* child */
#if defined(__linux__)
	setsid();
	if(grantpt(master) < 0) ERR("grantpt");
	if(unlockpt(master) < 0) ERR("unlockpt");
	if((name = ptsname(master)) == NULL) ERR("ptsname");
	if((slave = open(name, O_RDWR)) < 0) ERR("open");
	dup2(slave, 0);
	dup2(slave, 1);
	dup2(slave, 2);
#endif
	cui_dbg_close();
	execl("/bin/sh", "/bin/sh", "-i", (char *)NULL);
	exit(0);
}

cui
cui_terminal_new(cui parent, int x, int y, int w, int h)
{
	cui obj = cui_alloc(sizeof(struct cui_terminal));
	cui_terminal_init(obj, parent, x, y, w, h);
	return obj;
}

void
cui_terminal_init(cui obj, cui parent, int x, int y, int w, int h)
{
	cui_terminal p = (cui_terminal)obj;
	int evts;
	char *cmd = "stty erase ^H\r";

	cui_base_init(obj, parent, x, y, w, h);
	obj->flags |= CUI_FLG_CAN_FOCUS;

	p->term = cui_term_new(obj, 1, 0, w-2, h);
	p->in_key_mode = FALSE;

	p->fill_l = cui_fill_new(obj, 0, 0, 1, h, "|", CUI_ATTR_NORMAL);
	p->fill_r = cui_fill_new(obj, w-1, 0, 1, h-1, "|", CUI_ATTR_NORMAL);
	p->rszbox = cui_rszbox_new(obj);

	boot_sh(obj);
	if(write(p->pipe_w, cmd, strlen(cmd)) <0) ERR("write");

	evts = CUI_EVT_DRAW|CUI_EVT_KEY|CUI_EVT_RESIZE|CUI_EVT_TIMER|CUI_EVT_READABLE;
	cui_bind(obj, evts, cui_terminal_hdr, NULL);

	cui_readable_add(obj, p->pipe_r);
}

void
cui_terminal_in_key_mode_set(cui obj, int in_key_mode)
{
	int attr;

	cui_terminal p = (cui_terminal)obj;	
	if(p->in_key_mode == in_key_mode) return;

	p->in_key_mode = in_key_mode;
	attr = in_key_mode ? CUI_ATTR_REVERSE : CUI_ATTR_NORMAL;
	cui_fill_attr_set(p->fill_l, attr);
	cui_fill_attr_set(p->fill_r, attr);

	cui_timer_set(obj, TIMEOUT_MSEC, TRUE);
}

int
cui_terminal_hdr(cui obj, int evt, int val, void *prm)
{
	cui_terminal p = (cui_terminal)obj;
	char *s, buf[1024];
	int n;

	switch(evt){
	case CUI_EVT_DRAW:
		s = val == CUI_DRAW_FOCUS ? "/" : "|";
		cui_label_str_set(p->fill_l, s);
		cui_label_str_set(p->fill_r, s);
		return TRUE;

	case CUI_EVT_KEY:
		if(!p->in_key_mode){
			if(val == CUI_KEY_ENTER){
				cui_terminal_in_key_mode_set(obj, TRUE);
				return TRUE;
			}
			return FALSE;
		}
		/* in_key_mode */
		if(write(p->pipe_w, &val, 1) != 1) ERR("write");
		cui_timer_set(obj, TIMEOUT_MSEC, TRUE);
		return TRUE;

	case CUI_EVT_RESIZE:
		cui_wh_set(p->term, obj->w-2, obj->h);
		cui_wh_set(p->fill_l, 1, obj->h);
		cui_xywh_set(p->fill_r, obj->w-1, 0, 1, obj->h-1);
		cui_xy_set(p->rszbox, obj->w-1, obj->h-1);
		return TRUE;

	case CUI_EVT_TIMER:
		cui_terminal_in_key_mode_set(obj, FALSE);
		return TRUE;

	case CUI_EVT_READABLE:
		if((n = read(p->pipe_r, buf, 1024-1)) < 0){
			return TRUE;
		}
		buf[n] = '\0';
		cui_term_puts(p->term, buf);
		return TRUE;
	}
	return FALSE;
}

/* EOF */
